// Mr.Midi 2

// Hardware:
// ATmega8
// No crystal required (8MHz internal clock is sufficient)
// Uses UART for MIDI-IO
// SPI-IO plus extra Pin (CS) for SD-card
// Control buttons on PIND2-7
// LCD on PORTC

// Circuit Description:
/*
Take the circuit from Mr.Midi (the original Player/Recorder) 
available on www.mikrocontroller.net!
*/

// Software:
// MMC-Driver from alp_mp3 - Open Source Atmel AVR based MP3 Player (modified, though)
// My own MIDI-File-Parser (Format 0 only)
// Compiles with WinAVR plus AVRStudio 4

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include "lcd.h"
#include "mmc.h"
#include "spi.h"
#include "delay.h"
#include "fat16.h"
#include "types.h"

#ifdef __AVR_ATmega168__
#include "rc5.h"

//mega168
#define UDR UDR0
#define UCSRA UCSR0A
#define UCSRB UCSR0B
#define UCSRC UCSR0C
#define UBRRL UBRR0L
#define UBRRH UBRR0H
#define UDRE UDRE0
#define RXEN RXEN0
#define TXEN TXEN0
#define RXCIE RXCIE0
#define OCR2 OCR2A
#define TCCR0 TCCR0B
#define TCCR2 TCCR2B
#endif

#ifndef F_CPU
	#define F_CPU 8000000
#endif

// Config
#define BUFFERED_READ


#define RTC_PRESCALER	256
#define RTC_RELOAD_TIMER	250
#define RTC_SOFT_COUNT		125
#if F_CPU != (RTC_PRESCALER*RTC_RELOAD_TIMER*RTC_SOFT_COUNT)
	#error RTC_NOT_ADJUSTED
#endif

// Buttons
#define KEY_PORT PIND
#define KEY_PORT2 PINB

#ifdef __AVR_ATmega168__
	#define KEY_MASK 0xf8
	#define KEY_PLAY 0x80
#else
	#define KEY_MASK 0x7c
	#define KEY_PLAY 0x04
#endif

#define KEY_MASK2 0x03

#define KEY_STOP 0x08
#define KEY_REC 0x10
#define KEY_NEXT 0x20
#define KEY_LAST 0x40
#define KEY_LEFT 0x02
#define KEY_RIGHT 0x01

#define RC5_CODE_PLAY 0x35
#define RC5_CODE_STOP 0x36
#define RC5_CODE_REC 0x37
#define RC5_CODE_LEFT 0x32
#define RC5_CODE_RIGHT 0x34
#define RC5_CODE_NEXT 0x20
#define RC5_CODE_LAST 0x21
#define RC5_CODE_DELETE 0x31
#define RC5_CODE_MENU 0x1d

// States
#define STOP 0
#define PLAY 1
#define REC 2
#define ABORT_REC 3
#define DELETE 4
#define MENU 5

// Menu
#define MAX_MENU 5
#define M_DRUMCH 0
#define M_LYR 1
#define M_SSECT 2
#define M_MUTE 3
#define M_REP 4

// Flags
#define KEY_FLAG 0x01
#define DT_FLAG 0x02
#define OCCHNG_FLAG 0x04
#define IR_FLAG 0x08
#define IR_TGL 0x10
#define FAT_FLAG 0x20
#define MMCOK_FLAG 0x80

// MMC
#define START_SECTOR 528	// This is the sector number where the search for midi files begins (no FAT32 readings!)

// UART
#define UART_BAUD_RATE 31250
#define RX_BUF_SIZE 256	// lasts for at least 80ms on full MIDI data load (for *really* slow MMC write sector delays)

// Variables
u08 sharedmem[512];
u08 rx_buf[RX_BUF_SIZE], rx_rdp=0, rx_wrp=0;	// RX buffer
u08 flags=KEY_FLAG, dspd=0, state=STOP;	// Flags
u16 tempo;	// Tempo memory
u16 new_ocr1a;
u32 trk_len;	// Track length
u32 time=0, delta_time=0;	// Time
u08 time_min, time_sec, real_time=0, limit_count=RTC_SOFT_COUNT;
s08 speed, transpose;
u08 file_num=0, file_cnt=0;	// File
u16 *file_pos, last;
u08 ee_drumch=9, ee_lyr=1, ee_rep=2, menu_cnt, s_ch=0, gl_lyr=false;
u16 ee_ssect=START_SECTOR, ee_mute=0;
#ifdef ADC_TEMPO
	u16 ad_value=1024>>1;	// ADC
#endif
#ifdef BUFFERED_READ
	#define PLAYBUF_SIZE RX_BUF_SIZE
	#define PBS_EMPTY 1
	#define PBS_FULL 2
	#define playbuf rx_buf
	#define playbuf_rdp rx_rdp
	#define playbuf_wrp rx_wrp
	u08 playbuf_state=PBS_EMPTY;
#endif

extern struct fat_filedata_st fat_filedata;
extern u16 sect;


// Local Functions
void hw_init(void);
void menu_save(void);
void send_all_off(void);
void sendbyte(u08 byte);
void send_all_ch(u08 data);
u08 fetchbyte(void);
u08 getdata(void);
void writebyte(u08 b);
void writestring(u08 *buf, u08 len);
void checkbyte(u08 b);
u08 checkstring(u08 *buf, u08 len);
void write_varlen(u32 value);
void write_time(void);
u32 fetch_varlen(void);
void key_detect(void);
void debounce(u08 mode);
void wait_for_deltatime(void);
void print_time(void);
void clear_time(void);
void start_time(void);
void stop_time(void);
void print_main(void);
void print_menu(void);
void print_transpose(void);
void draw_mute(void);


//---------------------------------------------------

// Init MCU
void hw_init(void) {
//	u08 k;

	// Oscillator
#ifdef __AVR_ATmega8__
//	OSCCAL = 0xA1;
#endif

	// Ports
	DDRC = 0x3F;	// LCD
	PORTC = 0x00;
	lcd_init();
	lcd_string(DISP_MAIN, LINE_MAIN);
	lcd_string(DISP_BOOT, LINE_BOOT);

#ifdef ADC_TEMPO
	DDRB = 0x01;	// MMC & LCD enable
	PORTB = 0x3E;
#else
	DDRB = 0x00;	// MMC
	PORTB = 0x3F;
#endif

	DDRD = 0x02;	// Input-Port, auer TxD
	PORTD = 0xFF;	// Pull-Ups aktivieren

#ifdef ADC_TEMPO
	// ADC
	ADMUX = 0x40|5;
	ADCSRA = 0xe7;	// free running, start, prescl 128
#endif

	// Timer 0: Button Debounce
	// Timer 1: Delta Time Timer
	tempo = 2058;
	OCR1A = 2058;
	// Timer 2: RTC
	OCR2 = RTC_RELOAD_TIMER;

	// Interrupts
#ifdef __AVR_ATmega8__
	// mega8
	TIMSK = 0x51|0x80;	// Tmr 1&2 Compare Match Int. on, Overflow Timer 0 on
#endif
	// mega168
#ifdef __AVR_ATmega168__
//	TIMSK0 = 1;
	TIMSK1 = 2;
	TIMSK2 = 2;
#endif

	// UART
	// Hilfsmakro zur UBRR-Berechnung ("Formel" laut Datenblatt)
	#define UART_UBRR_CALC(BAUD_,FREQ_) ((FREQ_)/((BAUD_)*16L)-1)
	UBRRH = (uint8_t)(UART_UBRR_CALC(UART_BAUD_RATE,F_CPU)>>8);
	UBRRL = (uint8_t)UART_UBRR_CALC(UART_BAUD_RATE,F_CPU);
	UCSRB = (1<<RXEN)|(1<<TXEN)|(1<<RXCIE);//|(1<<UDRIE);

	// Midi device silence
//	send_all_off();

	// Read parameters from EEPROM
	if (eeprom_read_byte(0) != 0x27) {	// eeprom empty
		eeprom_write_byte(0, 0x27);
		menu_save();
	}
	else {	// load parameters
		ee_drumch = eeprom_read_byte(1);
		if (ee_drumch > 15)	// invalid channel
			ee_drumch = 10;
		ee_lyr = eeprom_read_byte(2);
		ee_ssect = eeprom_read_word(3);
		ee_mute = eeprom_read_word(5);
		ee_rep = eeprom_read_byte(7);
	}

	//Initialisierung der MMC/SD-Karte
	MMC_hw_init();
	spi_init();
#ifdef __AVR_ATmega168__
	rc5_init(RC5_ALL);
#endif
	sei();
	delay_ms(100);
	if (MMC_init()) {
		lcd_string(DISP_OK, LINE_OK);
		flags |= MMCOK_FLAG;
	}
	else
		lcd_string(DISP_CARDERR, LINE_CARDERR);
}


// Save data
void menu_save(void) {
	eeprom_write_byte(1, ee_drumch);	
	eeprom_write_byte(2, ee_lyr);
	eeprom_write_word(3, ee_ssect);
	eeprom_write_word(5, ee_mute);
	eeprom_write_byte(7, ee_rep);
}


//----------------------------------------------------------
// UART

// Send UART byte
void sendbyte(u08 byte) {
	loop_until_bit_is_set(UCSRA, UDRE);//while (!(UCSRA & (1<<UDRE)));
	UDR = byte;
}


// Sends to all channels
void send_all_ch(u08 data) {
    u08 i;
    for (i=0; i<16; i++) {
        sendbyte(0xB0+i);
        sendbyte(data);
        sendbyte(0x00);
    }
}


// All midi off
void send_all_off(void) {
    send_all_ch(0x7B);
    send_all_ch(0x79);
}


// Wait for UART RX
u08 getdata(void) {
	while (rx_rdp == rx_wrp && state == REC)
		key_detect();
	u08 b = rx_buf[rx_rdp++];
	rx_rdp &= RX_BUF_SIZE-1;
	return b;
}


//----------------------------------------------------------
// MMC IO

// Read from MMC
#ifdef BUFFERED_READ

// Abspielpuffer fllen
void fill_playbuf(void) {
	if (playbuf_state == PBS_FULL)
		return;
	while ((playbuf_rdp != playbuf_wrp || playbuf_state == PBS_EMPTY)) {
		playbuf[playbuf_wrp++] = mmc_fetch_byte();
		playbuf_wrp &= PLAYBUF_SIZE-1;
		playbuf_state = 0;
	}
	playbuf_state = PBS_FULL;
}

// Abspielpuffer initialisieren
void init_playbuf(void) {
	playbuf_wrp = 0;
	playbuf_rdp = 0;
	playbuf_state = PBS_EMPTY;
}

// Ein Byte vom Puffer holen
u08 read_playbuf(void) {
	u08 temp;

	if (playbuf_state == PBS_EMPTY)
		fill_playbuf();
	temp = playbuf[playbuf_rdp++];
	playbuf_rdp &= PLAYBUF_SIZE-1;
	if (playbuf_rdp == playbuf_wrp)
		playbuf_state = PBS_EMPTY;
	else
		playbuf_state = 0;
	return temp;
}

// Ein Byte von MMC lesen
u08 fetchbyte(void) {
	if (trk_len)
		trk_len--;
	return read_playbuf();
}
#else
// Ein Byte von MMC lesen
u08 fetchbyte(void) {
	trk_len--;
	return mmc_fetch_byte();
}
#endif


// Write byte to MMC
void writebyte(u08 b) {
	trk_len++;
	mmc_write_byte(b);
}


// Write string to MMC
void writestring(u08 *buf, u08 len) {
	u08 cnt=0;
	for (cnt=0; cnt<len; cnt++)
		mmc_write_byte(buf[cnt]);
}


// Read byte from MMC and verify value
void checkbyte(u08 b) {
	u08 temp = fetchbyte();
	if (b != temp)
	{
		lcd_string(DISP_FILEERR, LINE_FILEERR);
		state = STOP;
		delay_ms(1000);
	}
}


// Read string from MMC and verify content
u08 checkstring(u08 *buf, u08 len) {
	u08 b, cnt=0;

	for (cnt=0; cnt<len; cnt++) {
		b = fetchbyte();
		if (b != (u08)buf[cnt])
			return 0;
	}
	return 1;
}


//----------------------------------------------------------
// MIDI format

// Write Midi variable quantity to SD
void write_varlen(u32 value) {
	u32 buffer = value & 0x7f;
	while ((value >>= 7) > 0) {
		buffer <<= 8;
		buffer |= 0x80;
		buffer += (value & 0x7f);
	}
	while (1) {
		writebyte((u08)buffer);
		if (buffer & 0x80)
			buffer >>= 8;
		else
			break;
	}
}


// Write delta time to SD
void write_time(void) {
	write_varlen(time);
	time = 0;
}


// Read Midi variable quantity from SD
u32 fetch_varlen(void) {
	u32 v;
	u08 c;

	c = fetchbyte();
	v = c;
	if ((v & 0x80) > 0) {
		v &= 0x7F;
		do {
			c = fetchbyte();
			v = (v << 7) + (c & 0x7F);
		} while (c & 0x80);
	}
  	return v;
}


// Wait for delta time to expire
void wait_for_deltatime(void) {
	u32 dt;

	dt = fetch_varlen();
	if (time <= dt)
		delta_time = dt - time;
	else {
		delta_time = 0;
		time -= dt;
	}
	if (delta_time != 0) {
		flags |= DT_FLAG;
		while ((flags & DT_FLAG) && state == PLAY) {
			key_detect();
#ifdef BUFFERED_READ
			fill_playbuf();
#endif
		}
	}
}


// Tempo bestimmen
void calc_tempo(void) {
#ifdef ADC_TEMPO
	new_ocr1a = tempo * ad_value / 1024 + (tempo >> 1);
	flags |= OCCHNG_FLAG;
	TCNT2 = 0;	// TODO: Sicheres CTC benutzen!
	OCR2 = RTC_RELOAD_TIMER * ad_value / 1024 + RTC_RELOAD_TIMER / 2;
	lcd_setcur(TEMPO_X, TEMPO_Y);
	if (dspd) {
		lcd_number(ad_value / 5 + 98, 1);
		OCR2 = OCR2>>1;
	}
	else {
		lcd_number(ad_value / 10 + 49, 1);
	}
	lcd_data('%');
#else
	if (dspd) {
		TCNT1 = 0;
		dspd = 0;
	}
	new_ocr1a = tempo - (tempo>>4)*speed;
	flags |= OCCHNG_FLAG;
	lcd_setcur(TEMPO_X, TEMPO_Y);
	if (speed < 0) {
		lcd_data('-');
		lcd_nibble(-speed);
		//lcd_data('0'-speed);
	}
	else {
		lcd_data('+');
		lcd_nibble(speed);
		//lcd_data('0'+speed);
	}
/*	lcd_number(100+speed*10, 1);
	if (dspd==1) {
		lcd_number(200, 1);
		TCNT2 = 0;
		OCR2 = RTC_RELOAD_TIMER/2;
	}
	else if (dspd==2) {
		lcd_number(50, 1);
		TCNT2 = 0;
		OCR2 = RTC_RELOAD_TIMER;

	}
	else {
		lcd_number(100, 1);
		OCR2 = RTC_RELOAD_TIMER;
	}*/
	limit_count = RTC_SOFT_COUNT - RTC_SOFT_COUNT*speed/16;
	if (real_time >= limit_count)
		real_time = limit_count-1;
	//lcd_data('%');
#endif
}


//----------------------------------------------------------
// Buttons

// Entprellen
void debounce(u08 mode) {
	flags &= ~KEY_FLAG;
#ifdef __AVR_ATmega168__
	cli();
	WDTCSR = 0x18;	// Setup WDT-Init
	WDTCSR = 0x40|mode;	// WDIE=1->Interrupt mode; 32ms Timeout
	wdt_reset();
	sei();
#else
	TCNT0 = 0;
	TCCR0 = 5;	// clk/1024
#endif
}


// Check buttons
void key_detect(void) {
	static u08 ok=0, zif1=0xff;
	u08 k = ((u08)(~KEY_PORT)&KEY_MASK) | ((u08)(~KEY_PORT2)&KEY_MASK2);
	if ((k&KEY_NEXT) && (k&KEY_LAST))
		k = RC5_CODE_MENU;
	else if ((k&KEY_LEFT) && (k&KEY_RIGHT))
		k = RC5_CODE_DELETE;
	else if (k&KEY_STOP)
		k = RC5_CODE_STOP;
	else if (k&KEY_REC)
		k = RC5_CODE_REC;
	else if (k&KEY_LEFT)
		k = RC5_CODE_LEFT;
	else if (k&KEY_RIGHT)
		k = RC5_CODE_RIGHT;
	else if (k&KEY_NEXT)
		k = RC5_CODE_NEXT;
	else if (k&KEY_LAST)
		k = RC5_CODE_LAST;
	else if (k&KEY_PLAY)
		k = RC5_CODE_PLAY;

	if (!(flags & KEY_FLAG))
		return;
#ifdef ADC_TEMPO
	if (state == PLAY && TCCR1B && ad_value != ADC) {
		debounce();
		ad_value = ADC;
		calc_tempo();
	}
#endif
#ifdef __AVR_ATmega168__
		if (rc5.flip != -1) {
			k = rc5.code|0x80;
			if (k != ok || rc5.flip != ((flags>>4)&1) || (flags&IR_FLAG) == 0) {
				ok = k;
				k = 0;
			}
			flags |= IR_FLAG;
			if (rc5.flip)
				flags |= IR_TGL;
			else
				flags &= ~IR_TGL;
			rc5.flip = -1;
			debounce(3);
		}
#endif
	if (k != ok) {
		if (k != ok && !(flags&IR_FLAG)) {
			debounce(1);
			ok = k;
		}
		if (ok != 0) {
			k = ok & 0x7f;
			if (k <= 9 && file_cnt && (state == STOP)) {	// Zifferntaste
				if (file_cnt < 11) {
					if (k == 0)
						file_num=10;
					file_num = k-1;
					lcd_string(DISP_PLAY, LINE_PLAY);
					lcd_setcur(PLAYFNUM_X, PLAYFNUM_Y);
					lcd_number(file_num+1, 1);
					start_time();	// Start Time
					state = PLAY;
				}
				else {
					if (zif1 == 0xff) {
						zif1 = k;
						lcd_setcur(DWAHL_X, DWAHL_Y);
						lcd_data('-');
						lcd_data('>');
						lcd_data(k+'0');
					}
					else {
						file_num = k + zif1*10;
						if (file_num <= file_cnt && file_num > 0) {
							file_num--;
							lcd_string(DISP_PLAY, LINE_PLAY);
							lcd_setcur(PLAYFNUM_X, PLAYFNUM_Y);
							lcd_number(file_num+1, 1);
							start_time();	// Start Time
							state = PLAY;
						}
						lcd_setcur(DWAHL_X, DWAHL_Y);
						lcd_data(' ');
						lcd_data(' ');
						lcd_data(' ');
						zif1=0xff;
					}
				}
			}
			else if ((k == RC5_CODE_DELETE) && file_cnt && state == STOP) {	// Left&Right together->Delete last file
				state = DELETE;
			}
			else if ((k == RC5_CODE_MENU) && state == STOP) {	// Next&Last together->Enter Menu
				menu_cnt = 0;
				print_menu();
				state = MENU;
			}
			else if ((k == RC5_CODE_LAST)) {	// One file back
				if (state == STOP && file_cnt) {
					if (file_num == 0)
						file_num = file_cnt;
					file_num--;
					lcd_setcur(STOPFNUM_X, STOPFNUM_Y);
					lcd_number(file_num+1, 1);
					if (flags & FAT_FLAG) {
						fat_read_filedata(file_num);
						print_filename(NAME_LINE);
						lcd_setcur(STOPSECT_X, STOPSECT_Y);
						lcd_hex_u16(fat_filedata.startsect);
						print_filesize();
					}
					else {
						lcd_setcur(STOPSECT_X, STOPSECT_Y);
						lcd_hex_u16(file_pos[file_num]);
					}
				}
				else if (state == PLAY) {	// Transpose Down
					if (transpose > -24) {
						transpose--;
						print_transpose();
						send_all_off();
					}
				}
				else if (state == MENU) {	// Menu select
					menu_cnt++;
					if (menu_cnt == MAX_MENU)
						menu_cnt = 0;
					print_menu();
				}
			}
			else if ((k == RC5_CODE_NEXT)) {	// Next file
				if (state == STOP && file_cnt) {
					file_num++;
					if (file_num == file_cnt)
						file_num = 0;
					lcd_setcur(STOPFNUM_X, STOPFNUM_Y);
					lcd_number(file_num+1, 1);
					if (flags & FAT_FLAG) {
						fat_read_filedata(file_num);
						print_filename(NAME_LINE);
						lcd_setcur(STOPSECT_X, STOPSECT_Y);
						lcd_hex_u16(fat_filedata.startsect);
						print_filesize();
					}
					else {
						lcd_setcur(STOPSECT_X, STOPSECT_Y);
						lcd_hex_u16(file_pos[file_num]);
					}
				}
				else if (state == PLAY) {	// Transpose Up
					if (transpose < 24) {
						transpose++;
						print_transpose();
						send_all_off();
					}
				}
				else if (state == MENU) {	// Menu select
					menu_cnt--;
					if (menu_cnt >= MAX_MENU)
						menu_cnt = MAX_MENU-1;
					print_menu();
				}
			}
			else if (k == RC5_CODE_STOP) {	// Stop
				if (state == MENU) {	// Menu exit
					print_main();
					menu_save();
				}
				state = STOP;
			}
			else if ((k == RC5_CODE_PLAY) && file_cnt) {	// Play/Pause
				if (state == PLAY) {
					if (TCCR1B) {
						lcd_string(DISP_PAUSE, LINE_PAUSE);
						lcd_setcur(PLAYFNUM_X, PLAYFNUM_Y);
						lcd_number(file_num+1, 1);
						send_all_off();
						stop_time();	// Pause
					}
					else {
						lcd_string(DISP_PLAY, LINE_PLAY);
						lcd_setcur(PLAYFNUM_X, PLAYFNUM_Y);
						lcd_number(file_num+1, 1);
						start_time();	// Start Time
					}
				}
				else if (state == STOP && file_cnt) {
					state = PLAY;
				}
			}
			else if (k == RC5_CODE_REC) {	// Record
				if (state == STOP && flags & MMCOK_FLAG)
					state = REC;
				else if (state == REC)
					state = ABORT_REC;
			}
			else if (k == RC5_CODE_LEFT) {	// Left
				if (state == PLAY) {	// Slower
/*					dspd = 2;
					old_tempo = tempo;
					tempo <<= 1;*/
					if (speed > -15) {
						speed--;
						calc_tempo();
					}
				}
				else if (state == MENU) {	// Menu edit
					switch (menu_cnt) {
						case M_DRUMCH:
							if (ee_drumch > 0)
								ee_drumch--;
							break;
						case M_LYR:
							if (ee_lyr != 0)
								ee_lyr = 0;
							break;
						case M_SSECT:
							if (ee_ssect > 0)
								ee_ssect--;
							break;
						case M_MUTE:
							if (ee_mute & (1<<s_ch))
								ee_mute &= ~(1<<s_ch);
							else
								ee_mute |= 1<<s_ch;
							break;
						case M_REP:
							if (ee_rep > 0)
								ee_rep--;
							break;
					}
					print_menu();
				}
			}
			else if (k == RC5_CODE_RIGHT) {	// Right -> Faster
				if (state == PLAY) {	// Faster
					dspd = 1;
/*					old_tempo = tempo;
					tempo >>= 1;*/
					if (speed < 15) {
						speed++;
						calc_tempo();
					}
				}
				else if (state == MENU) {	// Menu edit
					switch (menu_cnt) {
						case M_DRUMCH:
							if (ee_drumch < 0x0f)
								ee_drumch++;
							break;
						case M_LYR:
							if (ee_lyr == 0)
								ee_lyr = 1;
							break;
						case M_SSECT:
							ee_ssect++;
							break;
						case M_MUTE:
							s_ch++;
							s_ch &= 0x0f;
							break;
						case M_REP:
							if (ee_rep<4)
								ee_rep++;
							break;
					}
					print_menu();
				}
			}
		}
	}
}


// Anzeige Hauptbildschirm
void print_main(void) {
	lcd_string(DISP_MAIN, LINE_MAIN);
	if (file_cnt) {
		lcd_string(DISP_SELECT, LINE_SELECT);
		lcd_setcur(STOPFNUM_X, STOPFNUM_Y);
		lcd_number(file_num+1, 1);
		if (flags & FAT_FLAG) {
			fat_read_filedata(file_num);
			print_filename(NAME_LINE);
			lcd_setcur(STOPSECT_X, STOPSECT_Y);
			lcd_hex_u16(fat_filedata.startsect);
			print_filesize();
		}
		else {
			lcd_setcur(STOPSECT_X, STOPSECT_Y);
			lcd_hex_u16(file_pos[file_num]);
		}
		lcd_setcur(FCNT_X, FCNT_Y);
		lcd_number(file_cnt, 1);
	}
	else if (flags & MMCOK_FLAG)
		lcd_string(DISP_NOFILE, LINE_NOFILE);
	else
		lcd_string(DISP_CARDERR, LINE_CARDERR);
}


// Anzeige Men
void print_menu(void) {
	lcd_string(DISP_MENU, LINE_MENU);
	lcd_string(DISP_MDRUM+menu_cnt, LINE_OPTION);
	switch (menu_cnt) {
		case M_DRUMCH:
			lcd_setcur(DRUMCH_X, LINE_OPTION);
			lcd_number(ee_drumch+1, 0);
			break;
		case M_LYR:
			lcd_setcur(LYR_X, LINE_OPTION);
			lcd_data(ee_lyr+'0');
			break;
		case M_SSECT:
			lcd_setcur(SSECT_X, LINE_OPTION);
			lcd_hex_u16(ee_ssect);
			break;
		case M_MUTE:
			draw_mute();
			break;
		case M_REP:
			lcd_setcur(REP_X, LINE_OPTION);
			lcd_data(ee_rep+'0');
			break;
	}
}


// Anzeige Transpose
void print_transpose(void) {
	lcd_setcur(TRANS_X, TRANS_Y);
	if (transpose < 0) {
		lcd_data('-');
		lcd_number(-transpose, 0);
	}
	else {
		lcd_data('+');
		lcd_number(transpose, 0);
	}
}


// Anzeige Muteleiste
void draw_mute() {
	uint8_t i;
	lcd_setcur(MUTE_X, MUTE_Y);
	for (i=0; i<16; i++) {
		if (ee_mute & (1<<i)) {
			if (i == s_ch)
				lcd_data('X');
			else
				lcd_data('x');
		}
		else {
			if (i == s_ch)
				lcd_data('O');
			else
				lcd_data('o');
		}
	}
}


//----------------------------------------------------------
// Time

// Zeit am Display anzeigen
void print_time(void) {
	lcd_setcur(TIME_X, TIME_Y);
	lcd_number(time_min, 0);
	lcd_data(':');
	lcd_number(time_sec, 0);
}


// Zeit zurcksetzen
void clear_time(void) {
	time_sec = 0;
	time_min = 0;
	real_time = 0;
	limit_count = RTC_SOFT_COUNT;
	print_time();
}


// Uhr starten
void start_time(void) {
	TCNT1 = 0;
	TCCR1B = 8+2;
	TCNT2 = 0;
	TCCR2 = 6+8;
}


// Uhr stoppen
void stop_time(void) {
	TCCR1B = 0;
	TCCR2 = 0;
}


//----------------------------------------------------------
// Main

void main(void) {
	u08 ev, len, adp=0, lyrx, lyry, trp=0, ch=0, ch_nact=0;
	u16 gl_timeset=96;
	u32 temp;

	file_pos = (u16 *)sharedmem;
	hw_init();

	if (flags & MMCOK_FLAG) {
		if (!fat_init()) {
			mmc_count_files(START_SECTOR);
		}
		else {
			flags |= FAT_FLAG;
			fat_count_files();
		}
		file_num = 0;
		print_main();
	}

	while (1) {
		switch (state) {
			case STOP:
				key_detect();
				break;

			case MENU:
				key_detect();
				break;

			case PLAY:
				// Wiedergabe starten
				lcd_string(DISP_PLAY, LINE_PLAY);	// Anzeige Wiedergabebildschirm
				lcd_string(DISP_TEMPO, LINE_TEMPO);
				lcd_setcur(PLAYFNUM_X, PLAYFNUM_Y);
				lcd_number(file_num+1, 1);
				clear_time();
				transpose = 0;
				lcd_setcur(TRANS_X, TRANS_Y);
				if (transpose < 0)
					lcd_data('-');
				else
					lcd_data('+');
				lcd_number(transpose, 0);

				if (flags & FAT_FLAG) {
					fat_openfile(FILEMODE_READ, file_num);
				}
				else {
					mmc_load_start(file_pos[file_num]);
				}
#ifdef BUFFERED_READ
				init_playbuf();
#endif
				if (!checkstring("MThd", 4)) {
					mmc_complete_read();
					lcd_string(DISP_FILEERR, LINE_FILEERR);
					delay_ms(1000);
					state = STOP;
					if (flags & FAT_FLAG)
						fat_closefile();
					print_main();
					break;
				}
				
				fetchbyte();
				checkbyte(0);	// Format 0 prfen
				checkbyte(0);
				checkbyte(6);
				checkbyte(0);
				checkbyte(0);
				checkbyte(0);
				checkbyte(1);
				gl_timeset = fetchbyte() << 8;	// Tempo lesen
				gl_timeset |= fetchbyte();

				TCNT1 = 0;
				// Tempo berechnen
				if (gl_timeset < 0x8000)	// Delta-Ticks
					tempo = 500000/gl_timeset;
				else				// SMPTE-Timeformat
					tempo = 1000000/(-(s08)(gl_timeset>>8))/(u08)gl_timeset;

				checkstring("MTrk", 4);	// Track lesen
			
				temp  = (u32)fetchbyte() << 24;	// Lnge lesen
				temp |= (u32)fetchbyte() << 16;
				temp |= (u32)fetchbyte() << 8;
				temp |= (u32)fetchbyte();
				trk_len = temp;
				time = 0;
				speed = 0;
				gl_lyr = false;

				lyrx = 0; lyry = 0;
			
				calc_tempo();
				start_time();	// Start Time
			
				while (state == PLAY) {
					if (trk_len == 0) {	// Tracklen wird heruntergezhlt. Bei 0 = Ende erreicht
						state = STOP;
						break;
					}
					wait_for_deltatime();
					if (state == STOP)
						break;
					ev = fetchbyte();
					if (ev == 0xFF) {	// META-EVENT
						ev = fetchbyte();
						switch (ev) {
							case 0x2f:	// End Of Track
								checkbyte(0);
								trk_len = 0;
								break;
							case 0x51:	// Tempo Change
								checkbyte(3);
								temp  = (u32)fetchbyte() << 16;
								temp |= (u32)fetchbyte() << 8;
								temp |= (u32)fetchbyte();
								TCNT1 = 0;
								tempo = temp/gl_timeset;
								/*if (dspd) {
									old_tempo = tempo;
									tempo >>= 1;
								}*/
								calc_tempo();
								break;
							//case 0x01:	// Text
							case 0x05:	// Lyrics
								if (ee_lyr) {
									len = 0;
									gl_lyr = true;
									temp = fetch_varlen();
									lcd_setcur(lyrx, lyry);
									while (temp > 0) {
										temp--;
										ev = fetchbyte();
										if (ev >= ' ' && ev <= '~') {
											lcd_data(ev);
											lyrx++;
											if (lyrx == LINESIZE) {
												lyrx = 0;
												lyry++;
												if (lyry == MAX_LINES)
													lyry = 0;
												lcd_string(DISP_BLANK, lyry);
												lcd_setcur(0, lyry);
											}
										}
									}
									break;
								}
							default:
								temp = fetch_varlen();
								while (temp > 0) {
									temp--;
									fetchbyte();
								}
						}
					}
					else if (ev == 0xf0) {	// SYSEX
						sendbyte(0xf0);
						temp = fetch_varlen();
						while (temp > 0) {
							temp--;
							ev = fetchbyte();
							sendbyte(ev);
						}
					}
					else if (ev >= 0x80) {	// MIDI-EVENT
						trp = false;
						//sendbyte(ev);
						ch_nact = 1;
						switch (ev) {
							case 0xf2:
								len = 2;
								break;
							case 0xf3:
								len = 1;
								break;
							default:
								ch = ev & 0x0f;
								ch_nact = 0;
								switch (ev & 0xf0) {
									case 0xf0:
										len = 0;
										break;
									case 0xc0:	// ProgramChange
									case 0xd0:	// ChannelAftertouch
										len = 1;
										break;
									case 0x80:	// NoteOff
									case 0x90:	// NoteOn
									case 0xa0:	// PolyphonicAftertouch
										trp = true;
									case 0xb0:	// ControlModeChange
									case 0xe0:	// PitchWheel
										len = 2;
										break;
									default:
										lcd_string(DISP_FILEERR, LINE_FILEERR);
										state = STOP;
										delay_ms(1000);
										//print_main();
										len = 0;
										trk_len = 0;
								}
						}
						adp = len;	// Lnge merken (fr Running Status)
						if (ch_nact || (ee_mute & ((uint16_t)1<<(uint16_t)ch)) == 0) {
							sendbyte(ev);
							if (trp && ch != ee_drumch) {
								sendbyte(fetchbyte()+transpose);
								sendbyte(fetchbyte());
							}
							else
							while (len > 0) {
								sendbyte(fetchbyte());	// Bytes durchschleifen
								len--;
							}
						}
						else {
							while (len > 0) {
								fetchbyte();	// Bytes ignorieren
								len--;
							}
						}
					}
					else {			// Running Status
						if (ch_nact || (ee_mute & ((uint16_t)1<<(uint16_t)ch)) == 0) {
							if (trp && ch != ee_drumch)
								sendbyte(ev+transpose);
							else
								sendbyte(ev);
							len = adp - 1;
							while (len > 0) {
								sendbyte(fetchbyte());
								len--;
							}
						}
						else {
							len = adp - 1;
							while (len > 0) {
								fetchbyte();	// Bytes ignorieren
								len--;
							}
						}
					}
				}
				stop_time();
				if (flags & FAT_FLAG)
					fat_closefile();
				else
					mmc_complete_read();
				send_all_off();
				// repeat modes:
				// 0=single song
				// 1=repeat single song
				// 2=all songs
				// 3=repeat all songs
				if (trk_len == 0) {
					if (ee_rep > 1) {
						if (file_num < file_cnt-1) {
							file_num++;
							state = PLAY;
						}
						else if (ee_rep == 3) {
							file_num = 0;
							state = PLAY;
						}
					}
					else if (ee_rep == 1) {
						state = PLAY;
					}
				}
				print_main();
				break;

			case REC:
				// Rekorder-Task
				if (flags & FAT_FLAG)
					fat_openfile(FILEMODE_WRITE, file_cnt);
				mmc_write_start(last+1);
			
				OCR1A = 500000/240;	// Fixes Tempo festlegen
				trk_len = 490;
				time = 0;
				rx_wrp = 0;	// Puffer leeren
				rx_rdp = 0;

				lcd_string(DISP_RECRD, LINE_RECRD);	// Bildschirm anzeigen
				lcd_setcur(RECFNUM_X, RECFNUM_Y);
				lcd_number(file_cnt+1, 1);
				lcd_string(DISP_BLANK, LINE_REC_BLANK);

				clear_time();

				while (state == REC) {
					ev = getdata();	// Auf Midi-Event warten
					if (state == STOP || state == ABORT_REC)
						break;
					write_time();
					if (!TCCR1B) {
						start_time();//TCCR1B = 8+2;	// Start Time
						lcd_string(DISP_REC, LINE_REC);
						lcd_setcur(RECFNUM_X, RECFNUM_Y);
						lcd_number(file_cnt+1, 1);
					}
					writebyte(ev);
			
					if (ev == 0xf0) {	// SysEx
						do {
							ev = getdata();
							writebyte(ev);
						} while (ev != 0xf7);
					}
					else if (ev < 0x80) {	// Running status
						len = adp-1;
						while (len > 0) {
							ev = getdata();
							writebyte(ev);
							len--;
						}
					}
					else {
						// MidiEvent
						switch (ev) {
						case 0xf2:
							len = 2;
							break;
						case 0xf3:
							len = 1;
							break;
						default:
							switch (ev & 0xf0) {
							case 0xf0:
								len = 0;
								break;
							case 0xc0:	// ProgramChange
							case 0xd0:	// ChannelAftertouch
								len = 1;
								break;
							case 0x80:	// NoteOff
							case 0x90:	// NoteOn
							case 0xa0:	// PolyphonicAftertouch
							case 0xb0:	// ControlModeChange
							case 0xe0:	// PitchWheel
								len = 2;
								break;
							default:
								len = 0;
							}
						}
						adp = len;
						while (len > 0) {
							ev = getdata();
							writebyte(ev);
							len--;
						}
					}
				}
				if (!TCCR1B || state == ABORT_REC) {
					stop_time();
					state = STOP;
					mmc_complete_write();
					print_main();
					break;
				}
				stop_time();

				writebyte(0);
				writebyte(0xff);
				writebyte(0x2f);
				writebyte(0);
				gl_timeset = sect + 1;	// Nchster freier Sektor
				mmc_complete_write();

				mmc_write_start(last);
		//		writestring("MThd", 4);
		//		mmc_write_byte(0xff);
		//		for (i=0; i<512-18-5-4; i++)
		//			mmc_write_byte(0xff);
				writestring("MThd", 4);
				mmc_write_byte(0);
				mmc_write_byte(0);
				mmc_write_byte(0);
				mmc_write_byte(6);
				mmc_write_byte(0);
				mmc_write_byte(0);
				mmc_write_byte(0);
				mmc_write_byte(1);
				mmc_write_byte(0);
				mmc_write_byte(240);	// Fixes Tempo im Header
				writestring("MTrk", 4);
				mmc_write_byte((u08)(trk_len >> 24));
				mmc_write_byte((u08)(trk_len >> 16));
				mmc_write_byte((u08)(trk_len >> 8));
				mmc_write_byte((u08)trk_len);
				mmc_write_byte(0);
				mmc_write_byte(0xff);
				mmc_write_byte(2);
				write_varlen(485);
				mmc_complete_write();

				if (flags & FAT_FLAG) {
					fat_filedata.startsect = last;
					//last = gl_timeset;
					fat_filedata.len = trk_len + 20;
					//fat_new_file();
					fat_closefile();
				}
				else {
					file_pos[file_cnt] = last;
					last = gl_timeset;
				}
				file_num = file_cnt;
				file_cnt++;

				print_main();
				break;

			case DELETE:
				state = STOP;
				if (file_cnt == 0)
					break;
				lcd_string(DISP_DELETE, LINE_DELETE);	// Bildschirm anzeigen
				file_cnt--;
				if (flags & FAT_FLAG) {
					fat_delete_file();
				}
				else {
					mmc_write_start(file_pos[file_cnt]);
					mmc_write_byte(0);
					mmc_complete_write();
				}
				if (file_num == file_cnt)
					file_num--;
				delay_ms(1000);
				print_main();
				break;
		}
	}
}


//----------------------------------------------------------
// IRQ

// Timer for delta times
ISR(SIG_OUTPUT_COMPARE1A) {
	time++;
	if ((flags & DT_FLAG) && time>=delta_time) {
		flags &= ~DT_FLAG;
		time = 0;
	}
	if (flags & OCCHNG_FLAG) {
		flags &= ~OCCHNG_FLAG;
		OCR1A = new_ocr1a;
	}
}


// Timer for real time clock
#ifdef __AVR_ATmega168__
  ISR(SIG_OUTPUT_COMPARE2A) {
#else
  ISR(SIG_OUTPUT_COMPARE2) {
#endif
	real_time++;
	if (real_time == limit_count) {
		real_time = 0;
		time_sec++;
		if (time_sec == 60) {
			time_sec = 0;
			time_min++;
		}
		if (gl_lyr == 0)
			print_time();
	}
}


#ifdef __AVR_ATmega168__
ISR(SIG_WATCHDOG_TIMEOUT) {
	wdt_disable();
	flags |= KEY_FLAG;
	if (flags & IR_FLAG && rc5.flip == -1)
		flags &= ~IR_FLAG;
}
#else
// Timer for button debounce
ISR(SIG_OVERFLOW0) {
	TCCR0 = 0;
	flags |= KEY_FLAG;
}
#endif


// UART RX
#ifdef __AVR_ATmega168__
  ISR(SIG_USART_RECV) {
#else
  ISR(SIG_UART_RECV) {
#endif
	static u08 tmp;
	tmp = UDR;
	if (state != REC || tmp == 0xf8 || tmp == 0xfe)	// Midi Clock und Synch ausblenden
		return;
	rx_buf[rx_wrp++] = tmp;
	rx_wrp &= RX_BUF_SIZE-1;
	if (rx_wrp == rx_rdp) {	// Buffer overflow!
		lcd_string(DISP_BUFERR, LINE_BUFERR);
		state = STOP;
		delay_ms(1000);
		print_main();
	}
}

